
def canfloat(num):
    """Check whether `num` can be a float or not.

    Args:
        num (Any): Instance of any type.

    Returns:
        bool: True if num can be a float, False otherwise.
    """
    try:
        float(num)
        return True
    except ValueError:
        return False


def canint(num):
    """Check whether `num` can be an integer or not.

    Args:
        num (Any): Instance of any type.

    Returns:
        bool: True if num can be an integer, False otherwise.
    """
    try:
        int(num)
        return True
    except ValueError:
        return False


def isstring(s):
    """Check whether `s` is a string or not.

    Args:
        s (Any): Instance of any type.

    Returns:
        bool: True if s is a string, False otherwise.
    """
    if isinstance(s, str):
        return True
    else:
        return False




def upper_case_first_letter_of_word(String: str, lower_case_words: list = []):
    """Get a string with your string transformed to have the firts letter in upper case.

    Args:
        String (str): Your string to transform. (not in place)
        lower_case_words (list, optional): List of words which should allways be in lower case. Defaults to [].

    Returns:
        string: AThe transformed string.
    """
    wörter = String.split(" ")
    _ws = []
    for w in wörter:
        dw = ""
        for iw, cw in enumerate(w):
            if iw == 0:
                dw += cw.upper()
            else:
                dw += cw
        if w.lower() in [w.lower() for w in lower_case_words]:
            dw = w.lower()
        _ws.append(dw)
    dn = " ".join(_ws)
    return dn


def upper_case_first_letter_of_words(List: list[str], lower_case_words: list = []):
    """Get a list with the strings transformed to have the firts letter in upper case in your list.

    Args:
        Liste (list[str]): Your list of strings to transform. (not in place)
        lower_case_words (list, optional): List of words which should allways be in lower case. Defaults to [].

    Returns:
        list: A list containing the transformed strings.
    """
    result = []
    for n in List:
        dn = upper_case_first_letter_of_word(n, lower_case_words)
        result.append(dn)
    return result



from .closest_furthest_value import (
    furthest_value_in_list,
    closest_value_in_list,
    furthest_value_in_dict,
    closest_value_in_dict,
)
from .dublicates import find_dublicates, remove_dublicates
from .percentage import (
    get_percentage_as_fitted_string,
    progress_printer,
    main_and_sub_progress_printer,
    ProgressBar,
)
from .sorter import sorted_dict
from .string_transformation import (
    upper_case_first_letter_of_word,
    upper_case_first_letter_of_words,
)
from .unpack import unpack_list
from .list_printer import pretty_print_list, pretty_print_nested_list

from .help_with_pickle import pickle_pack, pickle_unpack
from .help_with_json import dump_as_json, get_from_json

from .zipping import (
    make_zip_archive_with_shutil,
    unpack_zip_archive_with_shutil,
    zip_dir_with_zipfile,
    unzip_with_zipfile,
)

from .rounding import *
from .types import *
from .filesystem import (
    delete_empty_directories,
    get_disc_informations,
    move_and_integrate_directory,
    move_file,
    copy_file,
    get_directory_size,
    get_all_subdir_sizes,
    copied_paths_to_list,
)
from .stoppable_thread import ThreadWithExc, StoppableThread


__version__ = "0.0.17"





def round_relative_to_decimal(number: float, digits: int) -> str:
    """Round a number to the specified number of decimal trailing digits.
     Example 1: 0 with 2 digits: 0.00
     Example 2: 1.234 with 2 digits: 1.23

    Args:
        - number (float): Number to round.
        - digits (int): Decimal trailing digits.

    Returns:
        str: String representation of the rounded number.
    """
    number = round(number, digits)
    sn = str(number)
    if not "." in sn:
        sn += "."
    while len(sn.split(".")[-1]) < digits:
        sn = sn + "0"
    return sn


def round_significantly_std_notation(number: float, significant_digits: int) -> str:
    """Round a number to have only significant digits. Output is in standard notation.
     If the number is greater than ten to the power of the significant digits, the number is rounded to have 0 trailing digits.

    Args:
        number (float): Number to round.
        significant_digits (int): Significant digits.

    Returns:
        str: String representation of the rounded number in standard notation.
    """
    negative = False
    if number < 0:
        negative = True
        number *= -1
    snum = str(number)
    if not "e" in snum.lower():
        dpos = spos = snum.find(".")
        if dpos == -1:
            dpos = len(snum)
            snum += "."
        for i, c in enumerate(snum):
            if c != "." and c != "0":
                spos = i
                break
        e = dpos - spos
    else:
        e = int(snum.lower().split("e")[-1])
    num = number
    if e < 1:
        num = number / 10**e
        nstr = round_relative_to_decimal(num, significant_digits - 1)
        deci = e * (-1)
        nstr = "0" + "." + (deci - 1) * "0" + nstr.replace(".", "")
        if e == 0:
            nstr = nstr[:-1]
    elif e == 1:
        nstr = round_relative_to_decimal(number, significant_digits - 1)
    else:
        nstr = str(round(number, 0))
        if len(nstr) > significant_digits:
            nstr = nstr.split(".")[0]
        else:
            while len(nstr) - 1 < significant_digits:
                if "." not in nstr:
                    nstr += "."
                nstr += "0"
    if negative:
        nstr = "-" + nstr
    return nstr


def round_significantly_sci_notation(number: float, significant_digits: int, e_for_10: str = "E") -> str:
    """Round a number to have only significant digits. Output is in scientific notation with the n-th power of ten as En.

    Args:
        number (float): Number to round.
        significant_digits (int): Significant digits.

    Returns:
        str: String representation of the rounded number in scientific notation with the n-th power of ten as En.
    """
    negative = False
    if number < 0:
        negative = True
        number *= -1
    snum = str(number)
    if not "e" in snum.lower():
        dpos = spos = snum.find(".")
        if dpos == -1:
            dpos = len(snum)
        for i, c in enumerate(snum):
            if c != "." and c != "0":
                spos = i
                break
        e = dpos - spos
    else:
        e = int(snum.lower().split("e")[-1])
    estr = ""
    num = number
    if e <= 0:
        estr = f"{e_for_10}{e}"
        num = number / 10**e
        nstr = round_relative_to_decimal(num, significant_digits - 1) + estr
    else:
        e -= 1
        estr = f"{e_for_10}{e}"
        num = number / 10**e
        nstr = round_relative_to_decimal(num, significant_digits - 1) + estr
    if negative:
        nstr = "-" + nstr
    return nstr



from natsort import natsorted


def sorted_dict(dic: dict, for_key_not_value: bool = False):
    if for_key_not_value:
        dic_sorted = {k: v for k, v in natsorted(dic.items(), key=lambda item: item[0])}
    else:
        dic_sorted = {k: v for k, v in natsorted(dic.items(), key=lambda item: item[1])}
    return dic_sorted



import os
import shutil
from string import ascii_uppercase

from var_print import varp
import win32api
from clipboard import paste
from colorful_terminal import *
from .rounding import round_relative_to_decimal, round_significantly_std_notation
from .percentage import get_percentage_as_fitted_string


def delete_empty_directories(
    directory: str, full_tree: bool = False, reverb: bool = False
):
    """Delete empty directories

    Args:
        directory (str): mother directory to search in
        full_tree (bool, optional): if True all subdirectories will be effected. Defaults to False.
        reverb (bool, optional): print which paths were removed. Defaults to False.

    Returns:
        list[str]: removed directory paths
    """
    removed = []
    if full_tree:
        for root, dirs, files in os.walk(directory, topdown=False):
            for dir_name in dirs:
                dir_path = os.path.join(root, dir_name)
                if not os.listdir(dir_path):
                    os.rmdir(dir_path)
                    if reverb:
                        colored_print(f"Deleted empty directory: {Fore.YELLOW}{dir_path}")
                    removed.append(dir_path)
    else:
        for f in os.listdir(directory):
            fp = os.path.join(directory, f)
            if os.path.isdir(fp):
                if os.listdir(fp) == []:
                    os.rmdir(fp)
                    if reverb:
                        colored_print(f"Deleted empty directory: {Fore.YELLOW}{fp}")
                    removed.append(fp)
    return removed


def get_disc_informations(print_out: bool = True):
    """Get the information about your discs.

    Args:
        print_out (bool, optional): If True the string will be printed. Defaults to True.

    Returns:
        str: string with the information
    """
    first = True
    out = ""
    for c in ascii_uppercase:
        try:
            (
                name,
                seriennummer,
                maxfilenamelenght,
                sysflags,
                filesystemname,
            ) = win32api.GetVolumeInformation(f"{c}:\\")
            # total, used, free = shutil.disk_usage("/")
            total, used, free = shutil.disk_usage(f"{c}:\\")
            total, used, free = (
                round(total / 10**9, 2),
                round(used / 10**9, 2),
                round(free / 10**9, 2),
            )
            if not first:
                out += "\n"
            out += f"Hard disk information for {c}:\n"
            out += f"\t\tName:                " + name + "\n"
            out += f"\t\Serial number:        " + seriennummer + "\n"
            out += f"\t\tmax filename lenght: " + maxfilenamelenght + "\n"
            out += f"\t\tsys flags:           " + sysflags + "\n"
            out += f"\t\tfilesystem name:     " + filesystemname + "\n"
            out += f"\tMemory:\n"
            out += f"\t\tTotal:" + total, "GB\n"
            out += f"\t\tUsed: " + used, "GB\n"
            out += f"\t\tFree: " + free, "GB\n"
            first = False
        except:
            pass
    if print_out:
        print(out)
    return out


def move_file(
    filepath: str,
    targetpath: str,
    use_shutil_move: bool = True,
    copy_function: callable = shutil.copyfile,
):
    """Move the file into the target directory. Renames the file by adding ' (counter)' just before the file extension if the filename allready exists.

    Args:
        filepath (str): file to move
        targetpath (str): target directory
        use_shutil_move (bool): since shutil.move gave me trouble ... -> will use copy_function and os.remove
        copy_function (callable): used in shutil.moved or separately with os.remove
    """
    if os.path.isfile(targetpath):
        targetpath = os.path.dirname(targetpath)
    if not os.path.isdir(targetpath):
        os.makedirs(targetpath)
    content = os.listdir(targetpath)
    filename = os.path.basename(filepath)
    fn, fe = os.path.splitext(filename)
    counter = 2
    while filename in content:
        filename = fn + f" {counter}" + fe
        counter += 1
    nfp = os.path.join(targetpath, filename)
    if use_shutil_move: 
        shutil.move(filepath, nfp, shutil.copyfile)
    else:
        copy_function(filepath, nfp)
        os.remove(filepath)


def copy_file(
    filepath: str,
    targetpath: str,
    copy_function: callable = shutil.copyfile,
):
    """Copy the file into the target directory. Renames the file by adding ' (counter)' just before the file extension if the filename allready exists.

    Args:
        filepath (str): file to move
        targetpath (str): target directory
        copy_function (callable): shutil copy function
    """
    if os.path.isfile(targetpath):
        targetpath = os.path.dirname(targetpath)
    if not os.path.isdir(targetpath):
        os.makedirs(targetpath)
    content = os.listdir(targetpath)
    filename = os.path.basename(filepath)
    fn, fe = os.path.splitext(filename)
    counter = 2
    while filename in content:
        filename = fn + f" {counter}" + fe
        counter += 1
    nfp = os.path.join(targetpath, filename)
    copy_function(filepath, nfp)


def move_and_integrate_directory(
    dirpath: str,
    targetpath: str,
):
    """Move the content of the directory (dirpath) to the target path. Removes the dirpath if empty.

    Args:
        dirpath (str): source
        targetpath (str): target / destination

    Returns
        list[str]: paths that couldn't be moved
    """
    folderpath = os.path.normpath(dirpath)
    fails = []

    def try_move(fp: str):
        rel_path = os.path.relpath(fp, folderpath)
        structure = rel_path.split("\\")
        try:
            if not os.path.isdir(fp):
                npdp = os.path.join(targetpath, *structure[: len(structure) - 1])
                if not os.path.isdir(npdp):
                    os.makedirs(npdp)
            shutil.move(fp, targetpath)
        except:
            if os.path.isdir(fp):
                for n in os.listdir(fp):
                    try_move(os.path.join(fp, n))
            else:
                fails.append(fp)

    for f in os.listdir(folderpath):
        fp = os.path.join(folderpath, f)
        try_move(fp)
    try:
        os.rmdir(folderpath)
    except:
        pass
    return fails


def get_all_subdir_sizes(
    dirpath: str,
    unit: str = "GB",
    round_to: int = 2,
    sort_for_size: bool = False,
    print_it: bool = False,
    percentages: bool = True,
    with_sum: bool = True,
    as_string: bool = True,
):
    """Get a dictionary of the sizes of all the contents of the given directory.

    Args:
        dirpath (str): Directory to get the sizes of
        unit (str, optional): Size unit. Defaults to "GB".
        round_to (int, optional): Round size to. Defaults to 2.
        sort_for_size (bool, optional): Sort the dictionary for size (ascending). Defaults to False.
        print_it (bool, optional): Immediately print the dictionary using varp form var_print. Defaults to False.
        percentages (bool, optional): Adds the percentage to the size string. Defaults to True.
        with_sum (bool, optional): Adds the sum to the dictionary. Defaults to True.

    Raises:
        ValueError: if unit is not B / MB or GB or a number representing the size like 10**9 for GB

    Returns:
        dict[str, str]: Dictionary with the file names (keys) and the size as a string with the unit (possibly with the percentage)
    """
    if unit == "B":
        u = 1
    elif unit == "MB":
        u = 10**6
    elif unit == "GB":
        u = 10**9
    elif type(unit) == int or type(unit) == float:
        u = unit
    else:
        raise ValueError("unit not recognized")
    mp = dirpath
    summe = 0
    max_nachkomma = 0
    subdir_to_size = {}
    max_size = sum([os.path.getsize(os.path.join(mp, f)) for f in os.listdir(mp)]) / u
    pres = len(str(max_size).split(".")[0])
    for d in os.listdir(mp):
        dp = os.path.join(mp, d)
        size = os.path.getsize(dp) / u
        pras = len(str(size).split(".")[0])
        extr = pres - pras
        add_space = extr * " "
        summe += size
        size = round_relative_to_decimal(size, round_to).strip(".")
        if as_string:
            value = f"{add_space}{size} {unit}"
        else:
            value = size
        subdir_to_size[d] = value
        nachkomma = len(value.split(".")[-1])
        if nachkomma > max_nachkomma:
            max_nachkomma = nachkomma
    if sort_for_size:
        subdir_to_size = {
            k: v for (k, v) in sorted(subdir_to_size.items(), key=lambda item: item[1])
        }

    size = round_relative_to_decimal(summe, round_to).strip(".")
    if with_sum:
        if as_string:
            subdir_to_size["<Sum>"] = f"{size} {unit}"
        else:
            subdir_to_size["<Sum>"] = size

    if percentages:
        subdir_to_size_perc = {}
        for sd, size in subdir_to_size.items():
            if sd != "<Sum>":
                perc = get_percentage_as_fitted_string(
                    float(size.strip(" " + unit)), summe, 2
                )
                nachkomma = len(size.split(".")[-1])
                if nachkomma == len(size):
                    nachkomma = len(unit)
                spc = max_nachkomma - nachkomma + 4
                if as_string:
                    subdir_to_size_perc[sd] = size + " " * spc + perc
                else:
                    subdir_to_size_perc[sd] = size, perc
            else:
                subdir_to_size_perc[sd] = size
    if print_it:
        try:
            varp(subdir_to_size_perc)
        except:
            varp(subdir_to_size)
    return subdir_to_size


def get_directory_size(
    start_path=".", unit: str = None, round_to: int = 2, round_to_precision: bool = True, as_string: bool = True
):
    "unit can be none -> float or B / MB / GB -> str"
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(start_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # skip if it is symbolic link
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)
    if round_to_precision:
        rounding = round_significantly_std_notation
    else:
        rounding = round
    if unit == None:
        total_size = total_size
    elif unit == "B":
        total_size = rounding(total_size / 1, round_to)
        if as_string:
            total_size = f"{total_size} {unit}"
    elif unit == "MB":
        total_size = rounding(total_size / 10**6, round_to)
        if as_string:
            total_size = f"{total_size} {unit}"
    elif unit == "GB":
        total_size = rounding(total_size / 10**9, round_to)
        if as_string:
            total_size = f"{total_size} {unit}"
    else:
        raise ValueError("unit not recognized")

    return total_size


def get_file_size(
    file_path: str, unit: str = None, round_to: int = 2, round_to_precision: bool = True, as_string: bool = True
):
    "unit can be none -> float or B / MB / GB -> str"
    total_size = os.path.getsize(file_path)
    if round_to_precision:
        rounding = round_significantly_std_notation
    else:
        rounding = round
    if unit == None:
        total_size = total_size
    elif unit == "B":
        total_size = rounding(total_size / 1, round_to)
        if as_string:
            total_size = f"{total_size} {unit}"
    elif unit == "MB":
        total_size = rounding(total_size / 10**6, round_to)
        if as_string:
            total_size = f"{total_size} {unit}"
    elif unit == "GB":
        total_size = rounding(total_size / 10**9, round_to)
        if as_string:
            total_size = f"{total_size} {unit}"
    else:
        raise ValueError("unit not recognized")

    return total_size


def copied_paths_to_list():
    "path copied in Windows to a list"
    urls = paste().replace("\r", "").split("\n")
    urls = [u[1:-1] for u in urls]
    return urls


def copy_with_metadata(
    src_path, dest_path, integrate_files=True, skip_existing=True, rename_existing=True
):
    # Check if the source path exists
    if not os.path.exists(src_path):
        print(f"Source path '{src_path}' does not exist.")
        return

    # Check if the source path is a file or directory
    if os.path.isfile(src_path):
        # Check if the file already exists in the destination
        if skip_existing and os.path.exists(dest_path):
            print(f"File '{src_path}' already exists in '{dest_path}'. Skipping.")
        elif rename_existing and os.path.exists(dest_path):
            base, ext = os.path.splitext(os.path.basename(dest_path))
            i = 1
            while os.path.exists(dest_path):
                new_filename = f"{base} ({i}){ext}"
                dest_path = os.path.join(os.path.dirname(dest_path), new_filename)
                i += 1
            print(
                f"File '{src_path}' already exists in '{dest_path}'. Renaming to '{os.path.basename(dest_path)}'."
            )
            shutil.copy2(src_path, dest_path)
        else:
            # Copy the file with metadata
            shutil.copy2(src_path, dest_path)
    elif os.path.isdir(src_path):
        # Check if the destination directory exists
        if os.path.exists(dest_path):
            if integrate_files:
                # Integrate files if the destination directory exists
                for item in os.listdir(src_path):
                    s = os.path.join(src_path, item)
                    d = os.path.join(dest_path, item)
                    if os.path.isdir(s):
                        # Recursively copy subdirectories
                        copy_with_metadata(
                            s, d, integrate_files, skip_existing, rename_existing
                        )
                    else:
                        # Check if the file already exists in the destination
                        if skip_existing and os.path.exists(d):
                            print(f"File '{s}' already exists in '{d}'. Skipping.")
                        elif rename_existing and os.path.exists(d):
                            base, ext = os.path.splitext(item)
                            i = 1
                            while os.path.exists(d):
                                new_filename = f"{base} ({i}){ext}"
                                d = os.path.join(dest_path, new_filename)
                                i += 1
                            print(
                                f"File '{s}' already exists in '{d}'. Renaming to '{os.path.basename(d)}'."
                            )
                            shutil.copy2(s, d)
                        else:
                            # Copy individual files with metadata
                            shutil.copy2(s, d)
            else:
                print(
                    f"Destination directory '{dest_path}' already exists. Use integrate_files=True to integrate files."
                )
        else:
            # Copy the directory with metadata
            shutil.copytree(src_path, dest_path, copy_function=shutil.copy2)
    else:
        print(f"Source path '{src_path}' is neither a file nor a directory.")


def copy_without_metadata(
    src_path, dest_path, integrate_files=True, skip_existing=True, rename_existing=True
):
    # Check if the source path exists
    if not os.path.exists(src_path):
        print(f"Source path '{src_path}' does not exist.")
        return

    # Check if the source path is a file or directory
    if os.path.isfile(src_path):
        # Check if the file already exists in the destination
        if skip_existing and os.path.exists(dest_path):
            print(f"File '{src_path}' already exists in '{dest_path}'. Skipping.")
        elif rename_existing and os.path.exists(dest_path):
            base, ext = os.path.splitext(os.path.basename(dest_path))
            i = 1
            while os.path.exists(dest_path):
                new_filename = f"{base} ({i}){ext}"
                dest_path = os.path.join(os.path.dirname(dest_path), new_filename)
                i += 1
            print(
                f"File '{src_path}' already exists in '{dest_path}'. Renaming to '{os.path.basename(dest_path)}'."
            )
            shutil.copy(src_path, dest_path)
        else:
            # Copy the file without metadata
            shutil.copy(src_path, dest_path)
    elif os.path.isdir(src_path):
        # Check if the destination directory exists
        if os.path.exists(dest_path):
            if integrate_files:
                # Integrate files if the destination directory exists
                for item in os.listdir(src_path):
                    s = os.path.join(src_path, item)
                    d = os.path.join(dest_path, item)
                    if os.path.isdir(s):
                        # Recursively copy subdirectories
                        copy_without_metadata(
                            s, d, integrate_files, skip_existing, rename_existing
                        )
                    else:
                        # Check if the file already exists in the destination
                        if skip_existing and os.path.exists(d):
                            print(f"File '{s}' already exists in '{d}'. Skipping.")
                        elif rename_existing and os.path.exists(d):
                            base, ext = os.path.splitext(item)
                            i = 1
                            while os.path.exists(d):
                                new_filename = f"{base} ({i}){ext}"
                                d = os.path.join(dest_path, new_filename)
                                i += 1
                            print(
                                f"File '{s}' already exists in '{d}'. Renaming to '{os.path.basename(d)}'."
                            )
                            shutil.copy(s, d)
                        else:
                            # Copy individual files without metadata
                            shutil.copy(s, d)
            else:
                print(
                    f"Destination directory '{dest_path}' already exists. Use integrate_files=True to integrate files."
                )
        else:
            # Copy the directory without metadata
            shutil.copytree(src_path, dest_path)
    else:
        print(f"Source path '{src_path}' is neither a file nor a directory.")


def copy_without_metadata_using_copyfile(
    src_path, dest_path, integrate_files=True, skip_existing=True, rename_existing=True
):
    # Check if the source path exists
    if not os.path.exists(src_path):
        print(f"Source path '{src_path}' does not exist.")
        return

    # Check if the source path is a file or directory
    if os.path.isfile(src_path):
        # Check if the file already exists in the destination
        if skip_existing and os.path.exists(dest_path):
            print(f"File '{src_path}' already exists in '{dest_path}'. Skipping.")
        elif rename_existing and os.path.exists(dest_path):
            base, ext = os.path.splitext(os.path.basename(dest_path))
            i = 1
            while os.path.exists(dest_path):
                new_filename = f"{base} ({i}){ext}"
                dest_path = os.path.join(os.path.dirname(dest_path), new_filename)
                i += 1
            print(
                f"File '{src_path}' already exists in '{dest_path}'. Renaming to '{os.path.basename(dest_path)}'."
            )
            shutil.copyfile(src_path, dest_path)
        else:
            # Copy the file without metadata
            shutil.copyfile(src_path, dest_path)
    elif os.path.isdir(src_path):
        # Check if the destination directory exists
        if os.path.exists(dest_path):
            if integrate_files:
                # Integrate files if the destination directory exists
                for item in os.listdir(src_path):
                    s = os.path.join(src_path, item)
                    d = os.path.join(dest_path, item)
                    if os.path.isdir(s):
                        # Recursively copy subdirectories
                        copy_without_metadata(
                            s, d, integrate_files, skip_existing, rename_existing
                        )
                    else:
                        # Check if the file already exists in the destination
                        if skip_existing and os.path.exists(d):
                            print(f"File '{s}' already exists in '{d}'. Skipping.")
                        elif rename_existing and os.path.exists(d):
                            base, ext = os.path.splitext(item)
                            i = 1
                            while os.path.exists(d):
                                new_filename = f"{base} ({i}){ext}"
                                d = os.path.join(dest_path, new_filename)
                                i += 1
                            print(
                                f"File '{s}' already exists in '{d}'. Renaming to '{os.path.basename(d)}'."
                            )
                            shutil.copyfile(s, d)
                        else:
                            # Copy individual files without metadata
                            shutil.copyfile(s, d)
            else:
                print(
                    f"Destination directory '{dest_path}' already exists. Use integrate_files=True to integrate files."
                )
        else:
            # Copy the directory without metadata
            shutil.copytree(src_path, dest_path, copy_function=shutil.copyfile)
    else:
        print(f"Source path '{src_path}' is neither a file nor a directory.")


def delete_path(path):
    try:
        shutil.rmtree(path)
    except:
        os.remove(path)




def furthest_value_in_list(List: list, number: int | float) -> int | float:
    return max(List, key=lambda x: abs(x - number))


def closest_value_in_list(List: list, number: int | float) -> int | float:
    return min(List, key=lambda x: abs(x - number))


def furthest_value_in_dict(Dict: dict, number: int | float) -> int | float:
    return max(Dict, key=lambda x: abs(Dict[x] - number))


def closest_value_in_dict(Dict: dict, number: int | float) -> int | float:
    return min(Dict, key=lambda x: abs(Dict[x] - number))



import math
from threading import Thread
from time import sleep, time

import know_the_time
from colorful_terminal import Fore, TermAct, colored_print
from exception_details import print_exception_details

from .rounding import round_relative_to_decimal


def get_percentage_as_fitted_string(
    count: int, total: int, round_to: int = 2, with_percentage_symbol: bool = True
):
    if total == 0:
        count = 0
        perc = 100
    else:
        perc = count / total * 100
    perc = str(round_relative_to_decimal(perc, round_to))
    if with_percentage_symbol:
        perc += " %"
    return perc


def progress_printer(
    count: int, total: int, pre_string: str = "Progress: ", post_string: str = ""
):
    TermAct.clear_current_line_action()
    print(
        f"\r{pre_string}{str(count).rjust(len(str(total)))} / {total}    ({get_percentage_as_fitted_string(count, total)}){post_string}",
        end="",
    )
    if count == total:
        print()


def main_and_sub_progress_printer(
    maincount: int,
    maintotal: int,
    subcount: int,
    subtotal: int,
    pre_string: str = "Progress: ",
    mainpre_string: str = "Main-Progress: ",
    subpre_string: str = "Sub-Progress: ",
    post_string: str = "",
    mainpost_string: str = "",
    subpost_string: str = "",
):
    if maincount == 0 and subcount == 0:
        print(TermAct.hide_cursor(), end="")

    if post_string == "":
        lines = 3
    else:
        lines = 4 + post_string.count("\n")
    if maincount != 0 and subcount != 0:
        for i in range(lines):
            colored_print(TermAct.cursor_previous_line, end="")
    if len(mainpre_string) < len(subpre_string):
        mainpre_string = mainpre_string.ljust(len(subpre_string))
    elif len(mainpre_string) > len(subpre_string):
        subpre_string = subpre_string.ljust(len(mainpre_string))
    length = (
        len(str(subtotal))
        if len(str(subtotal)) > len(str(maintotal))
        else len(str(maintotal))
    )
    maintotal_str = str(maintotal).rjust(length)
    subtotal_str = str(subtotal).rjust(length)
    maincount_str = str(maincount).rjust(length)
    _subcount = subcount if subtotal != 0 else 0
    subcount_str = str(_subcount).rjust(length)

    try:
        # s = ""
        # s += f"{pre_string}" + "\n"
        # s += f"\r{mainpre_string}{maincount_str} / {maintotal_str}  ({get_percentage_as_fitted_string(maincount, maintotal)}){mainpost_string}" + "\n"
        # s += f"\r{subpre_string}{subcount_str} / {subtotal_str}  ({get_percentage_as_fitted_string(subcount, subtotal)}){subpost_string}" + "\n"
        # s += post_string + "\n"
        # print(s + "\r", end="")
        s = f"{pre_string}"
        TermAct.clear_current_line_action()
        print(s)
        s = f"\r{mainpre_string}{maincount_str} / {maintotal_str}  ({get_percentage_as_fitted_string(maincount, maintotal)}){mainpost_string}"
        TermAct.clear_current_line_action()
        print(s)
        s = f"\r{subpre_string}{subcount_str} / {subtotal_str}  ({get_percentage_as_fitted_string(subcount, subtotal)}){subpost_string}"
        TermAct.clear_current_line_action()
        print(s)
        s = post_string
        TermAct.clear_current_line_action()
        print(s)
    except Exception as e:
        print("\n" * 20)
        print_exception_details(e)
        print("\n" * 20)
    if maincount == maintotal and (subcount == subtotal or subtotal == 0):
        print(TermAct.show_cursor(), end="")
    else:
        print(TermAct.cursor_up * 4 + "\r", end="")


class ProgressBar:
    def __init__(
        self,
        total: float,
        prefix: str = "Progress: ",
        prefix_color: tuple = (127, 184, 0),
        suffix: str = "",
        suffix_color: tuple = (12, 99, 231),
        show_progress: bool = True,
        progress_color: tuple = (255, 234, 0),
        show_percentage: bool = True,
        percentage_color: tuple = (255, 180, 0),
        show_timer: bool = True,
        timer_color: tuple = (247, 127, 0),
        timer_prefix: str = "",
        timer_with_millisec: bool = False,
        show_delta_timer: bool = True,
        delta_timer_color: tuple = (246, 81, 29),
        delta_timer_prefix: str = "Δ: ",
        bar_type: str = "━",
        total_bar_length: int = 50,
        bar_color: tuple = (0, 166, 237),
        background_color: tuple = (77, 77, 77),
        vanish_with_finish: bool = False,
        auto_start: bool = True,
        indentation: int = 0,
        indentation_block: str = "    ",
        constant_output: bool = False,
        constant_output_rate: float = 0.1,
        precision: int = 5,
        spacing: str = "  "
    ) -> None:
        self.progress = 0
        self.total = total
        self.prefix = prefix
        self.prefix_color = prefix_color
        self.suffix = suffix
        self.suffix_color = suffix_color
        self.show_progress = show_progress
        self.progress_color = progress_color
        self.show_percentage = show_percentage
        self.percentage_color = percentage_color
        self.show_timer = show_timer
        self.timer_color = timer_color
        self.timer_prefix = timer_prefix
        self.timer_with_millisec = timer_with_millisec
        self.show_delta_timer = show_delta_timer
        self.delta_timer_color = delta_timer_color
        self.delta_timer_prefix = delta_timer_prefix
        self.bar_type = bar_type
        self.total_bar_length = total_bar_length
        self.bar_color = bar_color
        self.background_color = background_color
        self.vanish_with_finish = vanish_with_finish
        self.subprogresses = {}
        self.indentation = indentation
        self.indentation_block = indentation_block
        self.constant_output = constant_output
        self.constant_output_rate = constant_output_rate
        self.precision = precision
        self.spacing = spacing
        self.ratio = 0
        self.lines = 1
        self.finished = False
        self._stop_output = False
        self.starting_time = None
        self.last_updated = None

        if auto_start:
            self.update(0)
            if constant_output:
                Thread(target=self._constant_output, daemon=True).start()

    def get_progress_str(self):
        self.ratio = (
            round(self.progress / self.total, self.precision)
            if not self.total == 0
            else 0
        )
        bars_amount = math.floor(self.ratio * self.total_bar_length)
        no_bars = self.total_bar_length - bars_amount
        done_str = Fore.rgb(*self.bar_color) + bars_amount * self.bar_type
        not_done_str = Fore.rgb(*self.background_color) + no_bars * self.bar_type
        bar = done_str + not_done_str + self.spacing
        _suffix = ""
        if self.show_progress:
            _suffix += (
                Fore.rgb(*self.progress_color)
                + f"{round(self.progress, self.precision)} / {self.total}".rjust(2*len(str(self.total))+2)
                + self.spacing
            )        
        if self.show_percentage:
            _suffix += (
                Fore.rgb(*self.percentage_color)
                + f"{round_relative_to_decimal(self.ratio*100, 2)} %".rjust(7)
                + self.spacing
            )
        if self.show_timer:
            if not self.starting_time:
                self.starting_time = time()
            timer = know_the_time.get_time_delta_prettystring(
                self.starting_time,
                time(),
                include_milliseconds=self.timer_with_millisec,
            )
            _suffix += Fore.rgb(*self.timer_color) + self.timer_prefix + timer + self.spacing
        if self.show_delta_timer:
            if not self.last_updated:
                self.last_updated = time()
            delta_timer = know_the_time.get_time_delta_prettystring(
                self.last_updated,
                time(),
                include_milliseconds=self.timer_with_millisec,
            )
            _suffix += (
                Fore.rgb(*self.delta_timer_color)
                + self.delta_timer_prefix
                + delta_timer
                + self.spacing
            )
        _suffix += Fore.rgb(*self.suffix_color) + self.suffix
        main_bar = (
            self.indentation * self.indentation_block
            + Fore.rgb(*self.prefix_color)
            + self.prefix
            + bar
            + _suffix
        )
        main_bar = "\n".join(
            [line + TermAct.erase_in_line() for line in main_bar.splitlines()]
        )
        return main_bar

    def update(self, amount, name=None):
        if amount:
            self.last_updated = time()
        else:
            amount = 0
        if not name:
            self.progress += amount
        output = self.get_progress_str()

        for sub_name, sub_progress in self.subprogresses.items():
            sub_progress: ProgressBar
            if sub_name == name and not sub_progress.finished:
                sub_progress.progress += amount
            elif sub_name == name:
                sub_progress.progress = amount
                sub_progress.finished = False
            sub_bar = sub_progress.get_progress_str()
            if sub_name == name and sub_progress.ratio == 1:
                sub_progress.finished = True
                sub_progress.starting_time = None
            output += "\n" + sub_bar

        self.lines = len(output.splitlines())
        output += TermAct.cursor_up * (self.lines - 1)

        self.finished = (
            True
            if all([s.ratio == 1 or s.finished for s in self.subprogresses.values()])
            and self.ratio == 1
            else False
        )

        self.output = output

        if not self.constant_output:
            self.output_progress()

    def output_progress(self):
        colored_print(TermAct.hide_cursor() + self.output, end="\r")

        if self.finished == True:
            print(TermAct.show_cursor(), end="")
            print((self.lines - 1) * "\n")

            if self.vanish_with_finish:
                TermAct.clear_previous_line_action(self.lines)

    def add_subprogress(
        self,
        name: str,
        total: float,
        prefix: str = "Progress: ",
        prefix_color: tuple = (127, 184, 0),
        suffix: str = "",
        suffix_color: tuple = (12, 99, 231),
        show_progress: bool = True,
        progress_color: tuple = (255, 234, 0),
        show_percentage: bool = True,
        percentage_color: tuple = (255, 180, 0),
        show_timer: bool = True,
        timer_color: tuple = (247, 127, 0),
        timer_prefix: str = "",
        timer_with_millisec: bool = False,
        show_delta_timer: bool = True,
        delta_timer_color: tuple = (246, 81, 29),
        delta_timer_prefix: str = "Δ: ",
        bar_type: str = "━",
        total_bar_length: int = 50,
        bar_color: tuple = (0, 166, 237),
        background_color: tuple = (77, 77, 77),
        indentation: int = 0,
        indentation_block: str = "    ",
        constant_output: bool = False,
        constant_output_rate: float = 0.1,
        precision: int = 5,
        spacing: str = "  "
    ):
        progress = ProgressBar(
            total=total,
            prefix=prefix,
            prefix_color=prefix_color,
            suffix=suffix,
            suffix_color=suffix_color,
            show_progress=show_progress,
            progress_color=progress_color,
            show_percentage=show_percentage,
            percentage_color=percentage_color,
            show_timer=show_timer,
            timer_color=timer_color,
            timer_prefix=timer_prefix,
            timer_with_millisec=timer_with_millisec,
            show_delta_timer=show_delta_timer,
            delta_timer_color=delta_timer_color,
            delta_timer_prefix=delta_timer_prefix,
            bar_type=bar_type,
            total_bar_length=total_bar_length,
            bar_color=bar_color,
            background_color=background_color,
            vanish_with_finish=False,
            auto_start=False,
            indentation=indentation,
            indentation_block=indentation_block,
            constant_output=constant_output,
            constant_output_rate=constant_output_rate,
            precision=precision,
            spacing=spacing,
        )
        self.subprogresses[name] = progress
        return progress

    def _constant_output(self):
        while not self.finished and not self._stop_output:
            self.update(None)
            sleep(self.constant_output_rate)

    def stop_output(self):
        self._stop_output = True


if __name__ == "__main__":
    from random import randrange
    total = randrange(5, 10)
    total2 = randrange(5, 10)
    progress = ProgressBar(
        total,
        suffix="0 done",
        vanish_with_finish=False,
        timer_with_millisec=True,  # , constant_output=True
    )
    p2 = progress.add_subprogress(
        2,
        total,
        prefix="Subpprogress: ",
        suffix="0 done",
        timer_with_millisec=True,
        indentation=1,
    )

    for i in range(total):
        sleep(randrange(10) / 10)
        for j in range(total2):
            sleep(randrange(10) / 10)
            progress.update(1, 2)
            p2.suffix = str(j+1) + " done"
        progress.update(1)
        progress.suffix = str(i+1) + " done"



from typing import Iterable


def find_dublicates(List: Iterable):
    """Get a list of Dubs-objects. Dubs-objects have the attributes value, number_of_dublicates and indices.

    Will not find '1' as a dublicate of 1 since they are not the same...

    Args:
        List (Iterable): what you want to check

    Returns:
        list[Dubs]: Dubs-objects have the attributes value, number_of_dublicates and indices.
    """

    class Dubs:
        def __init__(self, value, number_of_dublicates, indices) -> None:
            self.value = value
            self.number_of_dublicates = number_of_dublicates
            self.indices = indices

    dublicates = []
    schon_drinnen = []
    for Element in List:
        number = List.count(Element)
        if number > 1:
            indices = [i for i, x in enumerate(List) if x == Element]
            if not any([x == Element for x in schon_drinnen]):
                obj = Dubs(Element, number, indices)
                dublicates.append(obj)
                schon_drinnen.append(Element)
    return dublicates


def remove_dublicates(List: list):
    """Remove dulicates in a list. Not in place!

    Args:
        List (list): what you want to check

    Returns:
        list: list without dublicates.
    """

    class Dubs:
        def __init__(self, value, number_of_dublicates, indices) -> None:
            self.value = value
            self.number_of_dublicates = number_of_dublicates
            self.indices = indices

    List_ = List + []
    dublicates = []
    schon_drinnen = []
    for Element in List_:
        number = List_.count(Element)
        if number > 1:
            indices = [i for i, x in enumerate(List_) if x == Element]
            if not any([x == Element for x in schon_drinnen]):
                obj = Dubs(Element, number, indices)
                dublicates.append(obj)
                schon_drinnen.append(Element)
    all_indices_nested = [i.indices[1:] for i in dublicates]
    all_indices = []
    for e in all_indices_nested:
        all_indices.extend(e)
    all_indices.sort(reverse=True)
    for i in all_indices:
        List_.pop(i)
    return List_



import os
import zipfile
from shutil import make_archive as shutil_make_archive
from shutil import unpack_archive as shutil_unpack_archive

from colorful_terminal import TermAct


def printProgressBar(
    iteration,
    total,
    prefix="",
    suffix="",
    decimals=1,
    length=100,
    fill="█",
    printEnd="\r",
):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
        printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + "-" * (length - filledLength)
    print(f"\r{prefix} |{bar}| {percent}% {suffix}", end=printEnd)
    # Print New Line on Complete
    if iteration == total:
        print()


def make_zip_archive_with_shutil(output_filename: str, dir_name: str):
    shutil_make_archive(output_filename, "zip", dir_name)


def unpack_zip_archive_with_shutil(filename: str, extract_dir: str):
    shutil_unpack_archive(filename, extract_dir, "zip")


def zip_dir_with_zipfile(folderpath: str, zip_path: str):
    """zip a complete directory, with progress reporting

    Args:
        folderpath (str): path to folder to zip
        zip_path (str): path for zip file with name
    """
    TermAct.hide_cursor_action()

    def zipdir(path, ziph):
        # ziph is zipfile handle
        allfiles = []
        for root, dirs, files in os.walk(path):
            allfiles += files
        len_files = len(allfiles)
        printProgressBar(
            0,
            len_files,
            prefix="Progress of zip file creation:",
            suffix="Complete",
            length=50,
        )
        index = 0
        for root, dirs, files in os.walk(path):
            for file in files:
                ziph.write(
                    os.path.join(root, file),
                    os.path.relpath(os.path.join(root, file), os.path.join(path, "..")),
                )
                printProgressBar(
                    index + 1,
                    len_files,
                    prefix="Progress of zip file creation:",
                    suffix="Complete",
                    length=50,
                )
                index += 1

    with zipfile.ZipFile(zip_path, "w", zipfile.ZIP_DEFLATED) as zipf:
        zipdir(folderpath, zipf)
    zipf.close()
    TermAct.show_cursor_action()


def unzip_with_zipfile(extract_dir: str, zip_path: str):
    """Extract a zipfile

    Args:
        extract_dir (str): root directory for the content
        zip_path (_type_): path to zip file
    """
    zfile = zipfile.ZipFile(zip_path)
    zfile.extractall(extract_dir)
    zfile.close()



"""
https://stackoverflow.com/questions/323972/is-there-any-way-to-kill-a-thread
Credit to Philippe F who based his work on Killable Threads by Tomer Filiba
"""

import threading, ctypes, inspect


def _async_raise(tid, exctype):
    """Raises an exception in the threads with id tid"""
    if not inspect.isclass(exctype):
        raise TypeError("Only types can be raised (not instances)")
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
        ctypes.c_long(tid), ctypes.py_object(exctype)
    )
    if res == 0:
        raise ValueError("invalid thread id")
    elif res != 1:
        # "if it returns a number greater than one, you're in trouble,
        # and you should call it again with exc=NULL to revert the effect"
        ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid), None)
        raise SystemError("PyThreadState_SetAsyncExc failed")


class ThreadWithExc(threading.Thread):
    """A thread class that supports raising an exception in the thread from
    another thread.
    """

    def _get_my_tid(self):
        """determines this (self's) thread id

        CAREFUL: this function is executed in the context of the caller
        thread, to get the identity of the thread represented by this
        instance.
        """
        if not self.is_alive():
            raise threading.ThreadError("the thread is not active")

        # do we have it cached?
        if hasattr(self, "_thread_id"):
            return self._thread_id

        # no, look for it in the _active dict
        for tid, tobj in threading._active.items():
            if tobj is self:
                self._thread_id = tid
                return tid

        # TODO: in python 2.6, there's a simpler way to do: self.ident

        raise AssertionError("could not determine the thread's id")

    def raiseExc(self, exctype=Exception):
        """Raises the given exception type in the context of this thread.

        If the thread is busy in a system call (time.sleep(),
        socket.accept(), ...), the exception is simply ignored.

        If you are sure that your exception should terminate the thread,
        one way to ensure that it works is:

            t = ThreadWithExc( ... )
            ...
            t.raiseExc( SomeException )
            while t.isAlive():
                time.sleep( 0.1 )
                t.raiseExc( SomeException )

        If the exception is to be caught by the thread, you need a way to
        check that your thread has caught it.

        CAREFUL: this function is executed in the context of the
        caller thread, to raise an exception in the context of the
        thread represented by this instance.
        """
        _async_raise(self._get_my_tid(), exctype)


class StoppableThread(threading.Thread):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition."""

    def __init__(self, *args, **kwargs):
        super(StoppableThread, self).__init__(*args, **kwargs)
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()



from copy import deepcopy
from varname import nameof
from colorful_terminal import *
from copy import deepcopy


def get_string_as_fitted_table(data: list, padding: int = 2, min_width: int = None, alignment: str = "left"):
    """Generates a string representation of the given table.

    Args:
        data (list): List of rows with rows as lists of column entries. e.g. `[["row1 col1", "row1 col2"], ["row2 col1", "row2 col2"]]`. If there are not enough columns, whitespace will fill the remaining cells.
        padding (int, optional): Space between columns = `padding * " "`. Defaults to 2.
        min_width (int, optional): Minimum width for columns. Defaults to None.
        alignment (str, optional): Text alignment. Can be "left", "center", or "right". Defaults to "left".

    Returns:
        str: String representation of the table.
    """
    dataset = deepcopy(data)
    row_lengths = [len(row) for row in dataset]
    cols = max(row_lengths)
    
    for row in dataset:
        while len(row) < cols:
            row.append("")
    
    col_widths = []
    for i in range(cols):
        col = [row[i] for row in dataset]
        col_width = max(len(str(word)) for word in col) + padding  # padding
        col_widths.append(col_width)
    
    s = ""
    for row in dataset:
        for index, word in enumerate(row):
            if min_width is not None and col_widths[index] < min_width:
                col_widths[index] = min_width
            
            if alignment == "left":
                s += str(word).ljust(col_widths[index])
            elif alignment == "center":
                s += str(word).center(col_widths[index])
            elif alignment == "right":
                s += str(word).rjust(col_widths[index])
                
        if row != dataset[-1]:
            s += "\n"
    
    return s


def pretty_print_list(
    List: list,
    print_list_name: bool = True,
    colored: bool = True,
    name_rgb=(255, 175, 0),
    value_rgb=(0, 255, 255),
    syntax_rgb=(200, 200, 200),
):
    """This function generates a visually appealing string representation of a given list.

    Args:
        List (list): The list to be printed.
        print_list_name (bool, optional): If True, prints the name of the list. Defaults to True.
        colored (bool, optional): If True, applies colors to enhance readability. Defaults to True.
        name_rgb (tuple, optional): RGB values for the color of the list name. Defaults to (255, 175, 0).
        value_rgb (tuple, optional): RGB values for the color of list values. Defaults to (0, 255, 255).
        syntax_rgb (tuple, optional): RGB values for the color of syntax elements. Defaults to (200, 200, 200).
    """
    if colored:
        n_color = Fore.rgb(*name_rgb)
        v_color = Fore.rgb(*value_rgb)
        s_color = Fore.rgb(*syntax_rgb)
        reset = Fore.RESET
    else:
        n_color = ""
        v_color = ""
        reset = ""
        s_color = ""
    if print_list_name:
        try:
            list_name = nameof(List, frame=2)
        except:
            list_name = "<name of list not found>"
        name = n_color + list_name + reset + s_color + " = [\n" + reset
        spacer = "\t"
    else:
        name = ""
        spacer = ""
    len_list = len(List)
    for i, v in enumerate(List):
        v = repr(v)
        ink = f"{spacer}{v_color}{v}{reset}"
        if i != len_list - 1 and print_list_name:
            ink += s_color + "," + reset
        name += ink + "\n"
    if print_list_name:
        name += s_color + "]" + reset
    colored_print(name)


def pretty_print_nested_list(
    List: list[list],
    print_list_name: bool = True,
    colored: bool = True,
    name_rgb=(255, 175, 0),
    value1_rgb=(0, 255, 255),
    value2_rgb=(255, 255, 0),
    syntax_rgb=(200, 200, 200),
):
    """This function generates a visually appealing string representation of a nested list.

    Args:
        List (list of lists): The nested list to be printed.
        print_list_name (bool, optional): If True, prints the name of the list. Defaults to True.
        colored (bool, optional): If True, applies colors to enhance readability. Defaults to True.
        name_rgb (tuple, optional): RGB values for the color of the list name. Defaults to (255, 175, 0).
        value1_rgb (tuple, optional): RGB values for the color of odd-indexed list values. Defaults to (0, 255, 255).
        value2_rgb (tuple, optional): RGB values for the color of even-indexed list values. Defaults to (255, 255, 0).
        syntax_rgb (tuple, optional): RGB values for the color of syntax elements. Defaults to (200, 200, 200).

    Raises:
        ValueError: If List is empty
    """
    try:
        name = nameof(List, frame=2)
    except:
        name = "<name of list not found>"
    if colored:
        n_color = Fore.rgb(*name_rgb)
        v1_color = Fore.rgb(*value1_rgb)
        v2_color = Fore.rgb(*value2_rgb)
        s_color = Fore.rgb(*syntax_rgb)
        reset = Fore.RESET
    else:
        n_color = ""
        v1_color = ""
        v2_color = ""
        reset = ""
        s_color = ""
    vcols = (v1_color, v2_color)
    spacer = "\t"

    data = List
    padding = 1

    if len(data) == 0:
        raise ValueError("List is empty!")
    dataset = deepcopy(data)
    row_lenghts = []
    for row in dataset:
        row_lenghts.append(len(row))
    cols = max(row_lenghts)
    for row in dataset:
        while len(row) < cols:
            row.append("")
    col_widths = []
    for i in range(cols):
        col = [row[i] for row in dataset]
        col_width = max(len(str(word)) for word in col) + padding  # padding
        col_widths.append(col_width)
    s = ""
    if print_list_name:
        s += n_color + name + reset + " = [\n"
    len_data = len(dataset)
    for row_index, row in enumerate(dataset):
        len_row = len(row)
        if print_list_name:
            s += spacer
        s += s_color + "[" + reset

        for index, word in enumerate(row):
            str_word = str(word)
            if type(word) == str:
                str_word = "'" + str_word + "'"
            if str_word != "''":
                if index != len_row - 1 and not all(
                    [w == "" for w in row[index + 1 :]]
                ):
                    str_word += s_color + "," + reset
                s += (
                    vcols[index % 2]
                    + str_word
                    + "".ljust(-len(str(word)) + col_widths[index])
                    + reset
                )

        s += s_color + TermAct.cursor_back + "]" + reset
        if row_index != len_data - 1:
            if print_list_name:
                s += s_color + "," + reset
            s += "\n"
    if print_list_name:
        s += s_color + "\n]" + reset
    colored_print(s)



import pickle
import os


def pickle_pack(data, path, append=False):
    if os.path.isfile(path) and append == True:
        org_data = pickle_unpack(path)
        org_type = type(org_data)
        if org_type in (tuple, list):
            if org_type == tuple:
                data = (*org_data, data)
            elif org_type == list:
                data = [*org_data, data]
        else:
            data = (org_data, data)
    with open(path, "wb") as f:
        pickle.dump(data, f)


def pickle_pack(data, path):
    with open(path, "wb") as f:
        pickle.dump(data, f)


def pickle_unpack(path):
    with open(path, "rb") as f:
        data = pickle.load(f)
    return data



import json


def dump_as_json(data, path):
    with open(path, "w") as f: json.dump(json.dumps(data), f)
    
def get_from_json(path):
    with open(path) as f: data = json.load(f)
    info = json.loads(data)
    return info


from typing import Iterable


def unpack_iterable(iterabl: list | tuple):
    """go from a nested iterable to one list containing every previously nested item; will not iterate on strings"""
    singles = []
    for l in iterabl:
        if type(l) != str and isinstance(l, Iterable):
            ls = unpack_iterable(l)
            singles.extend(ls)
        else:
            singles.append(l)
    return singles


def unpack_list(List: list | tuple):
    """go from a nested iterable to one list containing every previously nested item; will not iterate on strings\nis the same as unpack_iterable"""
    return unpack_iterable(List)



# following https://www.youtube.com/watch?v=tEFkHEKypLI&t=176s und https://www.youtube.com/watch?v=GIF3LaRqgXo 12:48 / 29:26

# cd -> current dir
# cd "C:\Users\Creed\OneDrive\Schul-Dokumente\Programmieren\Python\Code Sammlung\Packages\creating_easy_tasks"

# python setup.py sdist bdist_wheel

# install current package but not in site-packages -> link to this
# pip install -e .


# twine upload dist/*
# twine upload --skip-existing dist/*
# twine upload --verbose --skip-existing dist/*
# twine upload --verbose dist/*

from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

DESCRIPTION = (
    "Normal python class tasks like sorting, closet/furthest value, dublicates, ..."
)

# Setting up
setup(
    name="easy_tasks",
    version="0.0.17 ",
    author="André Herber",
    author_email="andre.herber.programming@gmail.com",
    # url="https://github.com/ICreedenI/xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=[
        "colorful_terminal",
        "exception_details",
        "natsort",
        "var_print",
        "pywin32",
        "clipboard",
        "varname",
        "know_the_time",
    ],
    keywords=["python"],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
    ],
)
